//////////////////////////////////////////////////////////////////////////
//                                                                      //
// pow.cpp -                                                            //
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:09p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

//  Ln(2) * 0x4000000000000000
const  ULONG DTLN2 = 0x2c5c85fdf473de6a;

#ifdef __cplusplus
extern "C" {
#endif

//   
//  Ln(1+1/2)*c, Ln(1+1/4)*c, Ln(1+1/8)*c, ...
//  c = 04000000000000000
extern ULONG _DTLN[64];

//    
//	    ln(x) = ln(~=2)+ln(2^(-1)) = ln(~=2)-ln(2)
extern ULONG _DTMLN[61];

//       ,   [0,1]
inline ULONG _MulFix0x64(ULONG M1, ULONG M2, int& NewExp){
	ULONG rez=0;
	while (!(M1 & 0x8000000000000000)){
		NewExp--;
		M1<<=1;
	}
	while (!(M2 & 0x8000000000000000)){
		NewExp--;
		M2<<=1;
	}
	while( M2>>=1 ){
        if (M1 & 0x8000000000000000)
            rez += M2;
        M1<<=1;
	}
    if (rez & 0x8000000000000000){
        rez>>=1;
		NewExp++;
    }
	return rez;
}


//    
//   -  CORDIC
inline ULONG SrvLog(double f, int& NewExp, int& OldSgn, int& Sgn){

	LONG Mnt;                           //   
	int Exp;                            //   
	ULONG Arg;                          //     

    //   double
	DstrDouble( f, Mnt, Exp, OldSgn);

	//  
	if ( Exp ==0x400 )	// -x, +inf, NAN:  NAN
		return 0x4000000000000001;

	if ( f == 0 )	// 0 -> +inf
		return 0x4000000000000000;

	Mnt <<= 10;
	ULONG tmp; //       
	LONG rez = 0;
	Arg= 0x4000000000000000; /*   1,  
								        */
    Sgn = 0;
	NewExp = 0;
	int i=0,j=0;

	tmp = Arg;

	if (Exp==-1){
		Mnt>>=1;
		while (i<=58){  
			i++;
			tmp -= (tmp>>(i+1)) + ((tmp>>(i))&1); 
			if (tmp>=Mnt){
				if (j==0){
					j=i;
					NewExp=-j+1;
					rez<<=j;
				}
				Arg = tmp;
				rez += _DTMLN[i-1]>>(i-j);
				tmp -= (tmp>>(i+1)) + ((tmp>>(i))&1); 
				if (tmp>=Mnt){
					Arg = tmp;
					rez += _DTMLN[i-1]>>(i-j);
				}
				else
					tmp = Arg;
			}
			else
				tmp = Arg;

		}
		Exp=0;
		Sgn=1;
	}
	else
		while (i<=58){  
			i++;
			tmp += tmp>>i; 
			if (tmp<=Mnt){
				Arg = tmp;
				rez += _DTLN[i-1];
			}
			else 
				tmp = Arg;
		}
	
	LONG tln2 = DTLN2;

	if (Exp<-1){
		rez = -rez;
		Exp = -Exp;
		Sgn=1;
	}

	i=0;
	LONG NewMnt = rez;
	while (Exp>0){  
		i++;
		if (Exp & 1){
			NewMnt += tln2;
			rez = NewMnt;
			rez >>=1;
			NewExp = i;
		}
		Exp>>=1;
		NewMnt>>=1;
	}
	if (rez & 0x4000000000000000)
		NewExp++;

	return rez;
}

//   
//   -  CORDIC
inline double Exp(ULONG Mnt, int Exp, int Sgn){

	int NewExp;
	ULONG Arg = Mnt;

	int i=-Exp;
    if (i>0){
        Arg>>=i;
        i=0;
    }
	//     () ,  floor 

    LONG tmp; //       
	ULONG rez = 0x4000000000000000; /*   1,  
								        */
	NewExp = 0;
	tmp = Arg;

	while (i<0){  //  
		i++;
		tmp -= DTLN2; 
		if (tmp>0){
			Arg = tmp;
			NewExp += 1 << (-i);
		}
		else //    - 
			tmp = Arg;
		tmp <<=1;
		Arg = tmp;
	}
	tmp >>=1;
	Arg = tmp;
	if (NewExp>0x3ff){
		if (Sgn==1){
			if(NewExp>0x44f){
				if (!_FPIEEE_Flags)
					_FPIEEE_Signal( _FPIEEE_Underflow ); 
				return ConstructDouble( 0, -0x3ff, 0, 0);
			}
		}
		else{
			if (!_FPIEEE_Flags)
				_FPIEEE_Signal( _FPIEEE_Overflow ); 
			return ConstructDouble( 0, 0x400, 0, 0);	
		}
	}
	if (Sgn==1){
		Arg = DTLN2 - Arg;
		tmp = Arg;
		NewExp = -(++NewExp);
	}
	while (i!=62){  //  
		tmp -= _DTLN[i++];  //    ,
                            //    ,  TLN[]
                            //  ,    
		if (tmp>0){
			rez += rez >> i;        //  - !
			Arg = tmp;
		}
		else                        //    - 
			tmp = Arg;
	}

	if (NewExp<-0x3fe){	            //  -  	
		rez >>= (-0x3ff - NewExp)+1;
		NewExp = -0x3ff;
	}

	rez >>=9;	                    //    , ...
	rez = (rez>>1) + (rez & 1);     //   ,  .

	if (rez == 0x0020000000000000){	
		rez = 0;
		NewExp++;
	}

	return ConstructDouble( rez, NewExp, 0, 0);		
}

// pow --
//
//           
//
//      --.
//      CORDIC,    .
double pow (double X, double Y){   // pow(x, y) = exp( y * ln(x) )
    int XExp, XSgn, LogSgn;
    LONG YMnt;
	int YExp, YSgn;

    //  ,   
    //  double Dbg_Res = log(X);
    //	DstrDouble( Dbg_Res, YMnt, YExp, YSgn);

    ULONG XMnt = SrvLog(X, XExp, XSgn, LogSgn );    // ln(x)
	if ( XMnt ==0 )
		return 1;

	int NewExp;
	ULONG Mul;
	DstrDouble( Y, YMnt, YExp, YSgn);

	//  
	if ( YExp ==0 && YMnt==0x0010000000000000 && YSgn ==0 )	// y = 1
		return X;

	if ( YExp == -0x3ff && YMnt==0 )	// y = 0
		return ConstructDouble( 0, 0, 0, 0);	

    if ( XExp ==0x400 )	
        if (XMnt == 0){             // X = inf
		    if (YSgn == 0)
                return X;
            else                    
                return ConstructDouble( 0, -0x3ff, XSgn, 0);
        }
        else{                       // X = NAN
            return X;
        }

    if ( XSgn ){    //   
        if ( YExp < 0 )    // NAN
    		return ConstructDouble( 1, 0x400, 0, 0);	
        else if ( YExp > 52){
            XSgn = 0;
        }
        else if ( (( YMnt>>(52-YExp) )<<(52-YExp)) != YMnt ){   // Y - , NAN
    		return ConstructDouble( 1, 0x400, 0, 0);	
        }
        else if ( (( YMnt>>(53-YExp) )<<(53-YExp)) == YMnt )   // Y - , |X|^Y
            XSgn = 0;
                                                               //      -(|X|^Y)
    }
    NewExp = XExp + YExp;
    Mul =  _MulFix0x64(XMnt, YMnt, NewExp)>>1;
    NewExp += 14;

    //  ,   
    //  Dbg_Res = log(X) * Y;
    //	DstrDouble( Dbg_Res, YMnt, XExp, XSgn);
    //    return exp (ConstructDouble( Mul>>9, NewExp-2, YSgn^LogSgn, 0));

    if (XSgn)
        return -Exp(Mul, NewExp, YSgn^LogSgn );
    else
        return Exp(Mul, NewExp, YSgn^LogSgn );
}

#ifdef __cplusplus
}
#endif
